@@ -1,5 +1,5 @@
-"""Different APIs that takes an :class:`Aito Client object <aito.client.AitoClient>` as the first argument
-
+"""
+Different APIs that takes an :class:`Aito Client object <aito.client.AitoClient>` as the first argument
 """
 
 import logging
@@ -8,6 +8,7 @@ import time
 from os import PathLike, unlink
 from pathlib import Path
 from typing import Dict, List, BinaryIO, Union, Tuple, Iterable, Optional
+import traceback
 
 import ndjson
 import requests as requestslib
@@ -16,12 +17,11 @@ import aito.client.requests as aito_requests
 import aito.client.responses as aito_responses
 from aito.client import AitoClient, RequestError
 from aito.schema import AitoDatabaseSchema, AitoTableSchema, AitoColumnTypeSchema
-from aito.utils._file_utils import gzip_file, check_file_is_gzipped
+from aito.utils._file_utils import gzip_file, check_file_is_gzipped, read_ndjson_gz_file
 from aito.utils.data_frame_handler import DataFrameHandler
 
 LOG = logging.getLogger('AitoAPI')
 
-
 def get_version(client: AitoClient) -> str:
     """get the aito instance version
 
@@ -255,25 +255,32 @@ def copy_table(client: AitoClient, table_name: str, copy_table_name: str, replac
     )
 
 
-def optimize_table(client: AitoClient, table_name):
+def optimize_table(client: AitoClient, table_name, use_job: bool = True):
     """`optimize <https://aito.ai/docs/api/#post-api-v1-data-table-optimize>`__
-    the specified table after uploading the data
+    the specified table after uploading the data. By default uses a job for the operation
 
     :param client: the AitoClient instance
     :type client: AitoClient
     :param table_name: the name of the table
     :type table_name: str
-    :return:
-    :rtype:
+    :param use_job: use a job to run the optimize
+    :type use_job: bool
     """
     try:
-        client.request(method='POST', endpoint=f'/api/v1/data/{table_name}/optimize', query={})
+        if use_job:
+            job_request(client,
+                        job_endpoint=f'/api/v1/jobs/data/{table_name}/optimize',
+                        query={},
+                        polling_time=5)
+
+        else:
+            client.request(method='POST', endpoint=f'/api/v1/data/{table_name}/optimize', query={})
+        LOG.info(f'table {table_name} optimized')
     except Exception as e:
         LOG.error(f'failed to optimize: {e}')
-    LOG.info(f'table {table_name} optimized')
+        traceback.print_tb(e.__traceback__)
 
-
-def delete_entries(client: AitoClient, query: Dict):
+def delete_entries(client: AitoClient, query: Dict, use_job: bool = True):
     """`Delete the entries <https://aito.ai/docs/api/#post-api-v1-data-delete>`__ according to the criteria
     given in the query
 
@@ -281,10 +288,22 @@ def delete_entries(client: AitoClient, query: Dict):
     :type client: AitoClient
     :param query: the query to describe the target table and filters for which entries to delete.
     :type query: Dict
+    :param use_job: use a job to run the optimize
+    :type use_job: bool
     """
-    client.request(request_obj=aito_requests.DeleteEntriesRequest(query=query))
-    LOG.info(f'entries deleted')
+    try:
+        if use_job:
+            job_request(client,
+                        job_endpoint=f'/api/v1/jobs/data/_delete',
+                        query=query,
+                        polling_time=5)
+        else:
+            client.request(request_obj=aito_requests.DeleteEntriesRequest(query=query))
 
+        LOG.info(f'entries deleted')
+    except Exception as e:
+        LOG.error(f'failed to delete: {e}')
+        traceback.print_tb(e.__traceback__)
 
 def upload_entries(
         client: AitoClient,
@@ -451,15 +470,39 @@ def poll_file_processing_status(client: AitoClient, table_name: str, session_id:
         time.sleep(polling_time)
 
 
+def _stream_entries_from_gzip(binary_file: BinaryIO) -> Iterable[Dict]:
+    """Stream entries from a gzipped ndjson file.
+
+    This is a generator that yields entries one by one from a gzipped ndjson file,
+    which is memory-efficient for large files.
+
+    :param binary_file: binary file object of a gzipped ndjson file
+    :type binary_file: BinaryIO
+    :yield: entries from the file
+    :rtype: Iterable[Dict]
+    """
+    import gzip
+    import json
+    with gzip.open(binary_file, 'rt', encoding='utf-8') as f:
+        for line in f:
+            line = line.strip()
+            if line:
+                yield json.loads(line)
+
+
 def upload_binary_file(
         client: AitoClient,
         table_name: str,
         binary_file: BinaryIO,
         polling_time: int = 10,
-        optimize_on_finished: bool = True
+        optimize_on_finished: bool = True,
+        batch_size: int = 1000
 ):
     """`upload a binary file object to a table <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__
 
+    For multitenant instances, this function uses streaming batch uploads instead of S3 file upload,
+    as S3 file upload is not supported in multitenant environments.
+
     .. note::
 
         requires the client to be setup with the READ-WRITE API key
@@ -470,23 +513,39 @@ def upload_binary_file(
     :type table_name: str
     :param binary_file: binary file object
     :type binary_file: BinaryIO
-    :param polling_time: polling wait time
+    :param polling_time: polling wait time (only used for non-multitenant S3 upload)
     :type polling_time: int
     :param optimize_on_finished: :func:`optimize_table` when finished uploading, defaults to True
     :type optimize_on_finished: bool
+    :param batch_size: batch size for streaming upload (only used for multitenant), defaults to 1000
+    :type batch_size: int
     """
     LOG.debug(f'uploading file object to table `{table_name}`...')
-    init_upload_resp = initiate_upload_file(client=client, table_name=table_name)
-    upload_binary_file_to_s3(initiate_upload_file_response=init_upload_resp, binary_file=binary_file)
-    upload_session_id = init_upload_resp['id']
-    trigger_file_processing(client=client, table_name=table_name, session_id=upload_session_id)
-    poll_file_processing_status(
-        client=client, table_name=table_name, session_id=upload_session_id, polling_time=polling_time
-    )
 
-    LOG.info(f'uploaded file object to table `{table_name}`')
-    if optimize_on_finished:
-        optimize_table(client, table_name)
+    if client.is_multitenant:
+        # For multitenant instances, use streaming batch upload
+        LOG.info(f'using streaming upload for multitenant instance')
+        entries = _stream_entries_from_gzip(binary_file)
+        upload_entries(
+            client=client,
+            table_name=table_name,
+            entries=entries,
+            batch_size=batch_size,
+            optimize_on_finished=optimize_on_finished
+        )
+    else:
+        # For non-multitenant instances, use S3 file upload
+        init_upload_resp = initiate_upload_file(client=client, table_name=table_name)
+        upload_binary_file_to_s3(initiate_upload_file_response=init_upload_resp, binary_file=binary_file)
+        upload_session_id = init_upload_resp['id']
+        trigger_file_processing(client=client, table_name=table_name, session_id=upload_session_id)
+        poll_file_processing_status(
+            client=client, table_name=table_name, session_id=upload_session_id, polling_time=polling_time
+        )
+
+        LOG.info(f'uploaded file object to table `{table_name}`')
+        if optimize_on_finished:
+            optimize_table(client, table_name)
 
 
 def upload_file(
@@ -494,10 +553,14 @@ def upload_file(
         table_name: str,
         file_path: PathLike,
         polling_time: int = 10,
-        optimize_on_finished: bool = True
+        optimize_on_finished: bool = True,
+        batch_size: int = 1000
 ):
     """`upload a file <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__ to the specfied table
 
+    For multitenant instances, this function uses streaming batch uploads instead of S3 file upload,
+    as S3 file upload is not supported in multitenant environments.
+
     .. note::
 
         requires the client to be setup with the READ-WRITE API key
@@ -508,10 +571,12 @@ def upload_file(
     :type table_name: str
     :param file_path: path to the file to be uploaded
     :type file_path: PathLike
-    :param polling_time: polling wait time
+    :param polling_time: polling wait time (only used for non-multitenant S3 upload)
     :type polling_time: int
     :param optimize_on_finished: :func:`optimize_table` when finished uploading, defaults to True
     :type optimize_on_finished: bool
+    :param batch_size: batch size for streaming upload (only used for multitenant), defaults to 1000
+    :type batch_size: int
     :raises ValueError: incorrect file extension, should be .ndjson.gz
     """
     if not check_file_is_gzipped(file_path):
@@ -522,15 +587,23 @@ def upload_file(
             table_name=table_name,
             binary_file=f,
             polling_time=polling_time,
-            optimize_on_finished=optimize_on_finished
+            optimize_on_finished=optimize_on_finished,
+            batch_size=batch_size
         )
 
 
 def quick_add_table(
-        client: AitoClient, input_file: Union[Path, PathLike], table_name: str = None, input_format: str = None
+        client: AitoClient,
+        input_file: Union[Path, PathLike],
+        table_name: str = None,
+        input_format: str = None,
+        batch_size: int = 1000
 ):
     """Create a table and upload a file to the table, using the default inferred schema
 
+    For multitenant instances, this function uses streaming batch uploads instead of S3 file upload,
+    as S3 file upload is not supported in multitenant environments.
+
     :param client: the AitoClient instance
     :type client: AitoClient
     :param input_file: path to the input file to be uploaded
@@ -539,6 +612,8 @@ def quick_add_table(
     :type table_name: Optional[str]
     :param input_format: specify the format of the input file, defaults to the input file extension
     :type input_format: Optional[str]
+    :param batch_size: batch size for streaming upload (only used for multitenant), defaults to 1000
+    :type batch_size: int
     """
     df_handler = DataFrameHandler()
 
@@ -567,7 +642,7 @@ def quick_add_table(
     create_table(client, table_name, inferred_schema)
 
     with open(converted_tmp_file.name, 'rb') as in_f:
-        upload_binary_file(client=client, table_name=table_name, binary_file=in_f)
+        upload_binary_file(client=client, table_name=table_name, binary_file=in_f, batch_size=batch_size)
     converted_tmp_file.close()
     unlink(converted_tmp_file.name)
 
@@ -585,7 +660,11 @@ def _create_job_request(
             raise TypeError("create_job() requires either 'request_obj' or 'job_endpoint' and 'query'")
     else:
         if isinstance(request_obj, aito_requests.QueryAPIRequest):
+            LOG.debug(f'Creating job from Query request {request_obj}')
             create_job_req = aito_requests.CreateJobRequest.from_query_api_request(request_obj=request_obj)
+        elif isinstance(request_obj, aito_requests.DataAPIRequest):
+            LOG.debug(f'Creating job from Data request {request_obj}')
+            create_job_req = aito_requests.CreateJobRequest.from_data_api_request(request_obj=request_obj)
         else:
             create_job_req = request_obj
     return create_job_req
