@@ -2,13 +2,14 @@
 
 Data structure for the Aito Database Schema
 """
-
+import re
 import itertools
 import logging
 from abc import abstractmethod, ABC
 from collections import Counter
 from csv import Sniffer, Error as csvError
 from typing import List, Iterable, Dict, Optional
+import re
 
 import pandas as pd
 from langdetect import detect_langs
@@ -17,7 +18,6 @@ from aito.utils._json_format import JsonFormat, JsonValidationError
 
 LOG = logging.getLogger('AitoSchema')
 
-
 class AitoSchema(JsonFormat, ABC):
     """The base class for Aito schema component
 
@@ -52,6 +52,11 @@ class AitoSchema(JsonFormat, ABC):
     def __eq__(self, other):
         return self._compare_type(other) and self._compare_properties(other)
 
+    table_name_pattern = r'[^\/\".$\r\n\s]+'
+    column_name_pattern = r'[^\/\".$\r\n\s]+'
+    column_link_pattern = f'{table_name_pattern}\.{column_name_pattern}'
+    uuid_pattern = r'[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{12}'
+
 
 class AitoAnalyzerSchema(AitoSchema, ABC):
     """the base class for `Aito Analyzer <https://aito.ai/docs/api/#schema-analyzer>`__
@@ -716,7 +721,7 @@ class DataSeriesProperties :
         else:
             self.target_aito_dtype = DataSeriesProperties.pandas_dtype_to_aito_dtype(pandas_dtype)
 
-            
+
     @classmethod
     def pandas_dtype_to_aito_dtype(cls, pandas_dtype) -> str:
         """ Converts a pandas data type into Aito data type
@@ -761,12 +766,15 @@ class DataSeriesProperties :
             raise Exception(f'failed to infer aito data type')
 
         return DataSeriesProperties(inferred_dtype, lower_bound, upper_bound)
-    
-        
+
+
 class AitoDataTypeSchema(AitoSchema, ABC):
     """The base class for Aito DataType"""
 
-    _supported_data_types = ["Boolean", "Decimal", "Int", "String", "Text"]
+    _supported_data_types = [
+        "Boolean", "Decimal", "Int", "String", "Text",
+        "Boolean[]", "Decimal[]", "Int[]", "String[]", "Text[]"
+    ]
 
     def __init__(self, aito_dtype: str):
         """
@@ -833,6 +841,24 @@ class AitoDataTypeSchema(AitoSchema, ABC):
         """
         return self._aito_dtype == 'Decimal'
 
+    @property
+    def is_array(self) -> bool:
+        """returns True if the data type is an array type
+
+        :rtype: bool
+        """
+        return self._aito_dtype.endswith('[]')
+
+    @property
+    def element_type(self) -> str:
+        """returns the element type for array types, or the type itself for non-array types
+
+        :rtype: str
+        """
+        if self.is_array:
+            return self._aito_dtype[:-2]  # Remove '[]' suffix
+        return self._aito_dtype
+
     def to_json_serializable(self) -> str:
         return self._aito_dtype
 
@@ -859,6 +885,17 @@ class AitoDataTypeSchema(AitoSchema, ABC):
             return AitoStringType()
         if obj == 'Text':
             return AitoTextType()
+        # Array types
+        if obj == 'Boolean[]':
+            return AitoBooleanArrayType()
+        if obj == 'Int[]':
+            return AitoIntArrayType()
+        if obj == 'Decimal[]':
+            return AitoDecimalArrayType()
+        if obj == 'String[]':
+            return AitoStringArrayType()
+        if obj == 'Text[]':
+            return AitoTextArrayType()
 
     @property
     def comparison_properties(self) -> Iterable[str]:
@@ -941,9 +978,59 @@ class AitoTextType(AitoDataTypeSchema):
         return str
 
 
+# Array types
+
+class AitoBooleanArrayType(AitoDataTypeSchema):
+    """Aito Boolean Array Type"""
+    def __init__(self):
+        super().__init__('Boolean[]')
+
+    def to_python_type(self):
+        return list
+
+
+class AitoIntArrayType(AitoDataTypeSchema):
+    """Aito Int Array Type"""
+    def __init__(self):
+        super().__init__('Int[]')
+
+    def to_python_type(self):
+        return list
+
+
+class AitoDecimalArrayType(AitoDataTypeSchema):
+    """Aito Decimal Array Type"""
+    def __init__(self):
+        super().__init__('Decimal[]')
+
+    def to_python_type(self):
+        return list
+
+
+class AitoStringArrayType(AitoDataTypeSchema):
+    """Aito String Array Type"""
+    def __init__(self):
+        super().__init__('String[]')
+
+    def to_python_type(self):
+        return list
+
+
+class AitoTextArrayType(AitoDataTypeSchema):
+    """Aito Text Array Type"""
+    def __init__(self):
+        super().__init__('Text[]')
+
+    def to_python_type(self):
+        return list
+
+
 class AitoColumnLinkSchema(AitoSchema):
     """Link to a column of another table"""
 
+    # Essentially a combination of the table- and column-regexes
+    _column_link_regex = f'^{AitoSchema.column_link_pattern}$'
+
     def __init__(self, table_name: str, column_name: str):
         """
 
@@ -952,6 +1039,11 @@ class AitoColumnLinkSchema(AitoSchema):
         :param column_name: the name of the linked column
         :type column_name: str
         """
+        if not bool(re.match(AitoDatabaseSchema._full_table_name_regex, table_name)):
+            raise ValueError(f'Linked table name in "{table_name}.{column_name}" is not valid')
+        if not bool(re.match(AitoTableSchema._column_name_regex, column_name)):
+            raise ValueError(f'Linked column name in "{table_name}.{column_name}" is not valid')
+
         self._table_name = table_name
         self._column_name = column_name
 
@@ -980,7 +1072,7 @@ class AitoColumnLinkSchema(AitoSchema):
 
     @classmethod
     def json_schema(cls):
-        return {'type': 'string', 'pattern': r'^[^/".$\r\n\s]+\.[^/".$\r\n\s]+$'}
+        return {'type': 'string', 'pattern': AitoColumnLinkSchema._column_link_regex}
 
     @classmethod
     def from_deserialized_object(cls, obj: str) -> "AitoColumnLinkSchema":
@@ -1247,12 +1339,22 @@ class AitoTableSchema(AitoSchema):
     ...     table_schema[col].nullable = False
     """
 
+    _column_name_regex = f'^{AitoSchema.column_name_pattern}$'
+
     def __init__(self, columns: Dict[str, AitoColumnTypeSchema]):
         """
 
         :param columns: a dictionary of the table's columns' name and schema
         :type columns: Dict[str, AitoColumnTypeSchema]
         """
+        invalid_columns = []
+        for key in columns.keys():
+            if not bool(re.match(AitoTableSchema._column_name_regex, key)):
+                invalid_columns.append(key)
+
+        if len(invalid_columns) > 0:
+            raise ValueError(f'Column names are not valid: {invalid_columns}')
+
         self._columns = columns
 
     @property
@@ -1301,6 +1403,10 @@ class AitoTableSchema(AitoSchema):
         """
         if not isinstance(column_name, str):
             raise TypeError('the name of the column must be of type string')
+
+        if not bool(re.match(AitoTableSchema._column_name_regex, column_name)):
+            raise ValueError(f'The column name {column_name} is not valid')
+
         if not isinstance(value, AitoColumnTypeSchema):
             raise TypeError('column schema must be of type AitoColumnTypeSchema')
         self._columns[column_name] = value
@@ -1330,6 +1436,7 @@ class AitoTableSchema(AitoSchema):
 
     @classmethod
     def json_schema(cls):
+        pattern = AitoTableSchema._column_name_regex
         return {
             'type': 'object',
             'properties': {
@@ -1337,7 +1444,7 @@ class AitoTableSchema(AitoSchema):
                 'columns': {
                     'type': 'object',
                     'minProperties': 1,
-                    'propertyNames': {'pattern': r'^[^/".$\r\n\s]+$'}
+                    'propertyNames': {'pattern': pattern}
                 }
             },
             'required': ['type', 'columns'],
@@ -1391,7 +1498,18 @@ class AitoDatabaseSchema(AitoSchema):
 
     Can be thought of as a dict-like container for :class:`.AitoTableSchema` objects
     """
+
+    _full_table_name_regex = f"^{AitoSchema.table_name_pattern}$"
+
     def __init__(self, tables: Dict[str, AitoTableSchema]):
+        invalid_tables = []
+        for key in tables.keys():
+            if not bool(re.match(AitoDatabaseSchema._full_table_name_regex, key)):
+                invalid_tables.append(key)
+
+        if len(invalid_tables) > 0:
+            raise ValueError(f'Column names are not valid: {invalid_tables}')
+
         self._tables = tables
 
     def _validate_links(self):
@@ -1453,7 +1571,11 @@ class AitoDatabaseSchema(AitoSchema):
         """update the schema of the specified table
         """
         if not isinstance(table_name, str):
-            raise TypeError('the name of the column must be of type string')
+            raise TypeError('the name of the table must be of type string')
+
+        if not bool(re.match(AitoDatabaseSchema._full_table_name_regex, table_name)):
+            raise ValueError(f'Table name {table_name} is not valid')
+
         if not isinstance(value, AitoTableSchema):
             raise TypeError('the table schema must be of type AitoTableSchema')
         self._tables[table_name] = value
@@ -1488,7 +1610,7 @@ class AitoDatabaseSchema(AitoSchema):
             'properties': {
                 'schema': {
                     'type': 'object',
-                    'propertyNames': {'pattern': r'^[^/".$\r\n\s]+$'}
+                    'propertyNames': {'pattern': AitoDatabaseSchema._full_table_name_regex}
                 }
             },
             'required': ['schema'],
